/*
Este arquivo é parte do programa XBMCMove



    XBMCMove é um software livre; você pode redistribui-lo e/ou 

    modifica-lo dentro dos termos da Licença Pública Geral GNU como 

    publicada pela Fundação do Software Livre (FSF); na versão 2 da 

    Licença, ou (na sua opnião) qualquer versão.



    Este programa é distribuido na esperança que possa ser  util, 

    mas SEM NENHUMA GARANTIA; sem uma garantia implicita de ADEQUAÇÂO a qualquer

    MERCADO ou APLICAÇÃO EM PARTICULAR. Veja a

    Licença Pública Geral GNU para maiores detalhes.



    Você deve ter recebido uma cópia da Licença Pública Geral GNU

    junto com este programa, se não, escreva para a Fundação do Software

    Livre(FSF) Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/**
 * @file
 * File:   kinect.h
 * Author: Diogo Luiz da Silva
 *
 * Arquivo contendo funções relativas ao dispositivo
 */
#include "callbacks.h"

/**
 *  Função para finalizar corretamente o dispositivo
 */
void CleanupExit() {
    if (NULL != g_pSessionManager) {
        delete g_pSessionManager;
        g_pSessionManager = NULL;
    }
    if (NULL != g_pCircle) {
        delete g_pCircle;
        g_pCircle = NULL;
    }
    if (NULL != g_pSwipeD) {
        delete g_pSwipeD;
        g_pSwipeD = NULL;
    }

    g_Context.Release();
    g_Context2.Release();
}

#define CHECK_RC(nRetVal, what)										\
	if (nRetVal != XN_STATUS_OK)									\
	{																\
		printf("[ERROR] %s failed: %s\n", what, xnGetStatusString(nRetVal));\
		/*return nRetVal;*/												\
	}

/**
 *  Procedimento para iniciar variaveis de gestos personalizados
 */
int initKinect() {

    // Configure
    rc = g_Context2.InitFromXmlFile(SAMPLE_XML_FILE, g_ScriptNode, &errors);
    if (rc == XN_STATUS_NO_NODE_PRESENT) {
        XnChar strError[1024];
        errors.ToString(strError, 1024);
        printf("%s\n", strError);
    } else if (rc != XN_STATUS_OK) {
        printf("[ERROR] Open failed: %s\n", xnGetStatusString(rc));
        return rc;
    }

    rc = g_Context2.FindExistingNode(XN_NODE_TYPE_DEPTH, g_DepthGenerator);
    CHECK_RC(rc, "Find depth generator");
    rc = g_Context2.FindExistingNode(XN_NODE_TYPE_USER, g_UserGenerator);
    if (rc != XN_STATUS_OK) {
        rc = g_UserGenerator.Create(g_Context2);
        CHECK_RC(rc, "Find user generator");
    }

    XnCallbackHandle hUserCallbacks, hCalibrationStart, hCalibrationComplete, hPoseDetected, hCalibrationInProgress, hPoseInProgress;
    if (!g_UserGenerator.IsCapabilitySupported(XN_CAPABILITY_SKELETON)) {
        printf("Supplied user generator doesn't support skeleton\n");
        //return 1;
    }
    rc = g_UserGenerator.RegisterUserCallbacks(User_NewUser, User_LostUser, NULL, hUserCallbacks);
    CHECK_RC(rc, "Register to user callbacks");
    rc = g_UserGenerator.GetSkeletonCap().RegisterToCalibrationStart(UserCalibration_CalibrationStart, NULL, hCalibrationStart);
    CHECK_RC(rc, "Register to calibration start");
    rc = g_UserGenerator.GetSkeletonCap().RegisterToCalibrationComplete(UserCalibration_CalibrationComplete, NULL, hCalibrationComplete);
    CHECK_RC(rc, "Register to calibration complete");

    if (g_UserGenerator.GetSkeletonCap().NeedPoseForCalibration()) {
        g_bNeedPose = TRUE;
        if (!g_UserGenerator.IsCapabilitySupported(XN_CAPABILITY_POSE_DETECTION)) {
            printf("Pose required, but not supported\n");
            //return 1;
        }
        rc = g_UserGenerator.GetPoseDetectionCap().RegisterToPoseDetected(UserPose_PoseDetected, NULL, hPoseDetected);
        CHECK_RC(rc, "Register to Pose Detected");
        g_UserGenerator.GetSkeletonCap().GetCalibrationPose(g_strPose);
    }

    g_UserGenerator.GetSkeletonCap().SetSkeletonProfile(XN_SKEL_PROFILE_ALL);
    return 0;
}


/**
 *  Função para pegar as coordenadas do esqueleto do usuário
 */
float getJointImgCoordinates(const SkeletonCapability &skeletonCapability, const XnUserID userId, const XnSkeletonJoint skeletonJoint, float *v) {
    XnVector3D projective;
    XnSkeletonJointPosition skeletonJointPosition;
    skeletonCapability.GetSkeletonJointPosition(userId, skeletonJoint, skeletonJointPosition);

    g_DepthGenerator.ConvertRealWorldToProjective(1, &skeletonJointPosition.position, &projective);

    v[0] = projective.X;
    v[1] = projective.Y;
    v[2] = projective.Z / 1000.0f;

    return skeletonJointPosition.fConfidence;
}

/**
 *  Procedimento para iniciar variaveis de gestos padrões do NITE
 */
int carregarKinect() {

    //Iniciar Configurações do kinect, separado para organizar
    rc = XN_STATUS_OK;

    // Configure
    rc = g_Context.InitFromXmlFile(SAMPLE_XML_FILE, g_ScriptNode, &errors);
    if (rc == XN_STATUS_NO_NODE_PRESENT) {
        XnChar strError[1024];
        errors.ToString(strError, 1024);
        CleanupExit();
        printf("%s\n", strError);
    } else if (rc != XN_STATUS_OK) {
        CleanupExit();
        printf("%s\n",SAMPLE_XML_FILE);
        printf("Open failed: %s\n", xnGetStatusString(rc));
        return rc;
    }


    // Create the gesture and hands generators
    rc = g_GestureGenerator.Create(g_Context);
    rc = g_HandsGenerator.Create(g_Context);

    if (rc != XN_STATUS_OK) {
        CleanupExit();
        //printf("Couldn't initialize from file: %s\n", xnGetStatusString(rc));
        return rc;
    }

    // Register to callbacks
    g_GestureGenerator.RegisterGestureCallbacks(Gesture_Recognized, Gesture_Process, NULL, h1);
    g_HandsGenerator.RegisterHandCallbacks(Hand_Create, Hand_Update, Hand_Destroy, NULL, h2);


    // Create and initialize point tracker
    g_pSessionManager = new XnVSessionManager();
    rc = g_pSessionManager->Initialize(&g_Context, "RaiseHand", NULL);
    if (rc != XN_STATUS_OK) {
        printf("[ERROR] Couldn't initialize the Session Manager: %s\n", xnGetStatusString(rc));
        CleanupExit();
    }

    g_pSessionManager->RegisterSession(NULL, &SessionStart, &SessionEnd);

    // Registrar e Reconhecer gesto: Circulo
    g_pCircle = new XnVCircleDetector;
    g_pCircle->RegisterCircle(NULL, &CircleCB);
    g_pCircle->RegisterNoCircle(NULL, &NoCircleCB);
    g_pCircle->RegisterPrimaryPointCreate(NULL, &Circle_PrimaryCreate);
    g_pCircle->RegisterPrimaryPointDestroy(NULL, &Circle_PrimaryDestroy);
    //g_pSessionManager->AddListener(g_pCircle);

    // Registrar e Reconhecer gesto: Wave
    wc = new XnVWaveDetector();
    wc->RegisterWave(NULL, OnWaveCB);
    g_pSessionManager->AddListener(wc);

    //Registrar e Reconhecer gesto: Swipe
    g_pSwipeD = new XnVSwipeDetector(true);
    g_pSwipeD->RegisterSwipeLeft(NULL, SwipeLeftCB);
    g_pSwipeD->RegisterSwipeRight(NULL, SwipeRightCB);
    g_pSwipeD->RegisterSwipeUp(NULL, SwipeUpCB);
    g_pSwipeD->RegisterSwipeDown(NULL, SwipeDownCB);
    g_pSwipeD->SetSteadyMaxVelocity(0.8);
    //printf("VAR=%f\n",g_pSwipeD->GetSteadyMaxStdDev());
    //g_pSwipeD->SetSteadyMaxStdDev(50);
    g_pSwipeD->SetSteadyDuration(500);
    g_pSwipeD->SetXAngleThreshold(75);
    g_pSwipeD->SetYAngleThreshold(75);
    g_pSessionManager->AddListener(g_pSwipeD);

    g_pTrackPad = new XnVSelectableSlider2D(g_TP_XDim, g_TP_YDim);

    // Add TrackPad to the point tracker
    g_TrackPadHandle = g_pSessionManager->AddListener(g_pTrackPad);

    // Register for the Hover event of the TrackPad
    g_nItemHoverHandle = g_pTrackPad->RegisterItemHover(NULL, &TrackPad_ItemHover);
    // Register for the Value Change event of the TrackPad
    g_nValueChangeHandle = g_pTrackPad->RegisterValueChange(NULL, &TrackPad_ValueChange);
    // Register for the Select event of the TrackPad
    g_nItemSelectHandle = g_pTrackPad->RegisterItemSelect(NULL, &TrackPad_ItemSelect);

    // Register for Input Start event of the TrackPad
    g_nPrimaryCreateHandle = g_pTrackPad->RegisterPrimaryPointCreate(NULL, &TrackPad_PrimaryCreate);
    // Register for Input Stop event of the TrackPad
    g_nPrimaryDestroyHandle = g_pTrackPad->RegisterPrimaryPointDestroy(NULL, &TrackPad_PrimaryDestroy);

    initKinect();

}

/**
 *  Procedimento para iniciar captura de gestos
 */
void iniciar() {

    //Começa a ouvir
    rc = g_Context.StartGeneratingAll();
    CHECK_RC(rc, "StartGeneratingAll");

    g_GestureGenerator.AddGesture("RaiseHand", NULL);
    //g_GestureGenerator.AddGesture("Click", NULL);

    rc = g_Context2.StartGeneratingAll();
    CHECK_RC(rc, "StartGeneratingAll2");

}

/**
 *  Função para pegar Status
 */
XnStatus getStatus() {
    return rc;
}

/**
 *  Procedimento para verificar se usuário fez o gesto para finalizar o XBMC
 *  Gesto: X com os braços
 */
void gestoFinal() {
    if (!gFechar) {
        if (lh[0] != rh[0]) {
            if (((rh[1] < (t[1])) && (lh[1] < (t[1]))) && (rh[0] < lh[0])) {
                fecharCont++;
                if (fecharCont >= 20) {
                    printf("[INFO] Gesto 'Fechar' detectado!\nFinalizando XBMC!\n");
                    system("killall xbmc");
                    system("killall xbmc.bin");
                    gFechar = true;
                }
            } else {
                fecharCont = 0;
            }
        }
    }
}

/**
 *  Procedimento para verificar se usuário fez o gesto para alterar o volume do XBMC
 *  Gesto: Duas mãos levantadas, se afastar as mãos volume aumenta, se aproximar as mãos volume diminui 
 */
void gestoVolume() {
    if (lh[0] != rh[0]) {
        if (((rh[1] < (t[1])) && (lh[1] < (t[1]))) && (rh[0] > lh[0])) {
            if (posInit[0] != 0) {
                if (volumeCont >= 20) {
                    if ((rh[0] - lh[0]) >= (distInit * 1.9)) {
                        printf("[INFO] Aumentar MASTER!\n");
                        teclar->simular(XK_plus);
                        volumeCont = 20;
                    } else {

                        if ((rh[0] - lh[0]) <= (distInit * 0.1)) {
                            printf("[INFO] Diminur MASTER!\n");
                            teclar->simular(XK_minus);
                            volumeCont = 20;
                        } else {
                            if (((rh[0] - lh[0]) > (distInit * 1.4)) && ((rh[0] - lh[0]) < (distInit * 1.9))) {
                                printf("[INFO] Aumentar bastante!\n");
                                teclar->simular(XK_plus);
                                volumeCont = 15;
                            } else {
                                if ((rh[0] - lh[0]) < (distInit * 0.6)) {
                                    printf("[INFO] Diminuir bastante!\n");
                                    teclar->simular(XK_minus);
                                    volumeCont = 15;
                                } else {
                                    if (((rh[0] - lh[0]) > (distInit * 0.6)) && ((rh[0] - lh[0]) < (distInit))) {
                                        printf("[INFO] Diminuir um pouco!\n");
                                        teclar->simular(XK_minus);
                                        volumeCont = 5;
                                    } else {
                                        printf("[INFO] Aumentar um pouco!\n");
                                        teclar->simular(XK_plus);
                                        volumeCont = 5;
                                    }
                                }
                            }
                        }
                    }
                } else {
                    volumeCont++;
                }

            } else {
                posInit[0] = lh[0];
                posInit[1] = rh[0];
                distInit = posInit[1] - posInit[0];
                volumeCont = 1;
            }
        } else {
            volumeCont = distInit = posInit[0] = posInit[1] = 0;
        }
    }
}

/**
 *  Procedimento para atualizar gestos personalizados
 */
void *updateC(void *threadid) {

    g_Context2.WaitOneUpdateAll(g_DepthGenerator);

    //g_Context2.WaitAnyUpdateAll();

    SkeletonCapability skelCap = g_UserGenerator.GetSkeletonCap();

    nUsers = g_UserGenerator.GetNumberOfUsers();
    g_UserGenerator.GetUsers(userIds, nUsers);
    id = userIds[0];

    getJointImgCoordinates(skelCap, id, XN_SKEL_HEAD, head); // Pega posição da cabeça
    //getJointImgCoordinates(skelCap, id, XN_SKEL_RIGHT_SHOULDER, ombroD); //Ombro Direito
    //getJointImgCoordinates(skelCap, id, XN_SKEL_LEFT_SHOULDER, ombroE); //Ombro Esquerdo
    //getJointImgCoordinates(skelCap, id, XN_SKEL_RIGHT_ELBOW, cotoveloD); // Cotovelo Direito
    //getJointImgCoordinates(skelCap, id, XN_SKEL_LEFT_ELBOW, cotoveloE); // Cotovelo Esquerdo
    getJointImgCoordinates(skelCap, id, XN_SKEL_TORSO, t);
    getJointImgCoordinates(skelCap, id, XN_SKEL_RIGHT_HAND, rh);
    getJointImgCoordinates(skelCap, id, XN_SKEL_LEFT_HAND, lh);

    gestoVolume();
    gestoFinal();

    pthread_exit(NULL);
}

/**
 *  Procedimento para atualizar gestos
 */
void *update(void *threadid) {
    pthread_t update2;
    long t2;

    while (rodando) {

        pthread_create(&update2, NULL, updateC, (void *) t2);
        pthread_join(update2, NULL); //Aguarda thread ser finalizada corretamente!
        //updateC();
        // Le os proximos dados disponiveis
        g_Context.WaitAnyUpdateAll();

        // Processa as informações
        g_pSessionManager->Update(&g_Context);

        //if (inSession) {

        //}
    }
    pthread_exit(NULL);
}


